import logging
logger = logging.getLogger(__name__)

import glob, os, sys, uuid
from threatspec import config, data, parser, reporter, threatmodel

class ThreatSpecApp():
    
    def __init__(self):
        self.threat_library = threatmodel.ThreatLibrary()
        self.control_library = threatmodel.ControlLibrary()
        self.component_library = threatmodel.ComponentLibrary()

        self.threatmodel = threatmodel.ThreatModel()
        self.threatmodel.threat_library = self.threat_library
        self.threatmodel.control_library = self.control_library
        self.threatmodel.component_library = self.component_library
        
        self.threatmodel.run_id = uuid.uuid4().hex
        logger.debug("Setting run id to {}".format(self.threatmodel.run_id))

        self.config = config.Config()
        self.parser = None
        self.reporter = None
        
        self.loaded_source_paths = {}
        self.loaded_library_paths = {}

    def parse_source(self, paths, parent):
        for config_path in paths:
            abs_path = data.abs_path(parent, config_path.path)
            if abs_path in self.loaded_source_paths:
                continue # Skip as we have already processed this path before
            self.loaded_source_paths[abs_path] = True # We've seen it now
            if data.is_threatspec_path(abs_path):
                logger.debug("Found threatspec.yaml, loading source configuration from {}".format(abs_path))
                new_config = config.Config()
                
                new_config_file = data.abs_path(abs_path, "threatspec.yaml")
                logger.debug("Validating configuration")
                (valid, error) = data.validate_yaml_file(new_config_file, os.path.join("data", "config_schema.json"))
                if not valid:
                    logger.error("Couldn't validate the configation file {}: {}".format(abs_path, error))
                    sys.exit(0)
         
                new_config.load(data.read_yaml(new_config_file))
                self.parse_source(new_config.paths, abs_path)
            else:
                for path in data.recurse_path(abs_path):
                    if data.path_ignored(path, config_path.ignore):
                        logger.debug("Skipping ignored file path: {}".format(path))
                        continue
                    logger.debug("Parsing source files in path {}".format(path))
                    if os.path.isfile(path):
                        self.parser = parser.SourceFileParser(self.threatmodel)
                        self.parser.parse_file(path)
                        """
                        if os.path.splitext(filename)[1].lower() in [".json", ".yaml"]:
                            self.parser = parser.YamlFileParser(self.threatmodel)
                            self.parser.parse_file(filename)
                        else:
                            self.parser = parser.SourceFileParser(self.threatmodel)
                            self.parser.parse_file(filename)
                        """

    def load_threat_model(self, path):
        filename = data.abs_path(path, "threatmodel", "threatmodel.json")
        
        logger.debug("Validating {}".format(filename))
        (valid, error) = data.validate_yaml_file(filename, os.path.join("data", "threatmodel_schema.json"))
        if not valid:
            logger.error("Couldn't validate the threat model file {}: {}".format(filename, error))
            sys.exit(0)
            
        try:
            self.threatmodel.load(data.read_json(filename))
            logger.debug("Loaded threat model from {}".format(filename))
        except FileNotFoundError:
            pass

    def save_threat_model(self):
        data.write_json_pretty(self.threatmodel.save(), data.cwd(), "threatmodel", "threatmodel.json") # TODO: Unhardcode
    
    def load_threat_library(self, path):
        filename = data.abs_path(path, "threatmodel", "threats.json")
        
        logger.debug("Validating {}".format(filename))
        (valid, error) = data.validate_yaml_file(filename, os.path.join("data", "threats_schema.json"))
        if not valid:
            logger.error("Couldn't validate the threat library file {}: {}".format(filename, error))
            sys.exit(0)
            
        try:
            self.threat_library.load(data.read_json(filename))
            logger.debug("Loaded threat library from {}".format(filename))
        except FileNotFoundError:
            pass

    def load_control_library(self, path):
        filename = data.abs_path(path, "threatmodel", "controls.json")

        logger.debug("Validating {}".format(filename))
        (valid, error) = data.validate_yaml_file(filename, os.path.join("data", "controls_schema.json"))
        if not valid:
            logger.error("Couldn't validate the control library file {}: {}".format(filename, error))
            sys.exit(0)
            
        try:
            self.control_library.load(data.read_json(filename))
            logger.debug("Loaded control library from path {}".format(filename))
        except FileNotFoundError:
            pass

    def load_component_library(self, path):
        filename = data.abs_path(path, "threatmodel", "components.json")
        
        logger.debug("Validating {}".format(filename))
        (valid, error) = data.validate_yaml_file(filename, os.path.join("data", "components_schema.json"))
        if not valid:
            logger.error("Couldn't validate the components library file {}: {}".format(filename, error))
            sys.exit(0)
            
        try:
            self.component_library.load(data.read_json(filename))
            logger.debug("Loaded component library from path {}".format(filename))
        except FileNotFoundError:
            pass

    def load_threat_library_data_from_path(self, paths, parent):
        for config_path in paths:
            abs_path = data.abs_path(parent, config_path.path)
            if abs_path in self.loaded_library_paths:
                continue # Skip as we've seen this path before
            self.loaded_library_paths[abs_path] = True # We've seen it now
            if data.is_threatspec_path(abs_path):
                logger.debug("Found threatspec.yaml, loading library from {}".format(abs_path))
                self.load_threat_library(abs_path)
                self.load_control_library(abs_path)
                self.load_component_library(abs_path)
                
                new_config_file = data.abs_path(abs_path, "threatspec.yaml")
                new_config = config.Config()
                
                logger.debug("Validating {}".format(new_config_file))
                (valid, error) = data.validate_yaml_file(new_config_file, os.path.join("data", "config_schema.json"))
                if not valid:
                    logger.error("Couldn't validate the configation file {}: {}".format("threatspec.yaml", error))
                    sys.exit(0)
         
                new_config.load(data.read_yaml(new_config_file))
                self.load_threat_library_data_from_path(new_config.paths, abs_path)

    def load_threat_library_data(self):
        self.load_threat_library_data_from_path(self.config.paths, data.cwd())

    def save_threat_library_data(self):
        data.write_json_pretty(self.threat_library.save(self.threatmodel.run_id), data.cwd(), "threatmodel", "threats.json") # TODO: Unhardcode
        data.write_json_pretty(self.control_library.save(self.threatmodel.run_id), data.cwd(), "threatmodel", "controls.json") # TODO: Unhardcode
        data.write_json_pretty(self.component_library.save(self.threatmodel.run_id), data.cwd(), "threatmodel", "components.json")

    def load_threat_model_data_from_path(self, paths):
        self.load_threat_model(paths)

    def load_threat_model_data(self):
        self.load_threat_model_data_from_path(data.cwd())
        for path in self.config.paths:
            if os.path.dirname(path.path) == data.cwd():
                continue
            base_path = data.glob_to_root(path.path)
            if data.is_threatspec_path(base_path):
                self.load_threat_model_data_from_path(base_path)

    def save_threat_model_data(self):
        self.save_threat_model()

    def generate_report(self):
        self.reporter = reporter.MarkdownReporter(self.config.project, self.threatmodel)
        self.reporter.generate()
        
    def init(self):
        logger.info("Initialising threatspec...")

        logger.debug("Creating default configuration file")
        try:
            data.copy_pkg_file(os.path.join("data","default_config.yaml"), "threatspec.yaml")
        except FileExistsError as e:
            logger.error("Configuration file already exists, it looks like threatspec has already been initiated here.")
            sys.exit(0)

        config_file = data.abs_path(data.cwd(), "threatspec.yaml")
        logger.debug("Validating {}".format(config_file))
        (valid, error) = data.validate_yaml_file(config_file, os.path.join("data", "config_schema.json"))
        if not valid:
            logger.error("Couldn't validate the configation file {}: {}".format(config_file, error))
            sys.exit(0)
            
        logger.debug("Loading configuration")
        self.config.load(data.read_yaml(config_file))

        logger.debug("Creating directories")
        try:
            data.create_directories(["threatmodel"])
        except IOError as e:
            logger.error("Failed to create directories: {}".format(str(e)))
            raise
        logger.info("""
Threatspec has been initialised. You can now configure the project in this
repository by editing the following file:

    threatspec.yaml.
        """)

    def run(self):
        logger.info("Running threatspec...")
            
        config_path = data.abs_path(data.cwd(), "threatspec.yaml")
        
        logger.debug("Validating configuration")
        (valid, error) = data.validate_yaml_file(config_path, os.path.join("data", "config_schema.json"))
        if not valid:
            logger.error("Couldn't validate the configation file {}: {}".format("threatspec.yaml", error))
            sys.exit(0)
        self.config.load(data.read_yaml(config_path))

        self.load_threat_library_data()
        self.parse_source(self.config.paths, data.cwd())
        self.save_threat_library_data()
        self.save_threat_model_data()

        logger.info("""
Threatspec has been run against the source files. The following threat mode file
has been created and contains the mitigations, acceptances, connections etc. for
the project:

    threatmodel/threatmodel.json

The following library files have also been create:

    threatmodel/threats.json threatmodel/controls.json threatmodel/components.json
        """)

    def report(self):
        logger.info("Generating report...")

        logger.debug("Validating configuration")
        (valid, error) = data.validate_yaml_file("threatspec.yaml", os.path.join("data", "config_schema.json"))
        if not valid:
            logger.error("Couldn't validate the configation file {}: {}".format("threatspec.yaml", error))
            sys.exit(0)
         
        logger.debug("Loading configuration from threatspec.yaml")
        self.config.load(data.read_yaml("threatspec.yaml"))

        self.load_threat_library_data()

        self.load_threat_model_data()

        self.generate_report()

        logger.info("""
The following threat model markdown report has been created:

    ThreatModel.md

The following visualisation image used in the report has also been created:

    ThreatModel.gv.png
        """)



import re, uuid
from threatspec import data, threatmodel, config
from pprint import pprint
from graphviz import Digraph

class Reporter():
    def __init__(self, project: config.Project, threatmodel: threatmodel.ThreatModel):
        self.project = project
        self.threatmodel = threatmodel

        self.components = {}
        self.component_pairs: {}

    def random_id(self):
        return uuid.uuid4().hex

    def to_id(self, text):
        if text.endswith("/"):
            text += "root"
        return "#" + re.sub('[^a-z0-9_]+', '_', text.strip().lower().replace('-','')).strip('_')

    def parse_component_paths(self):
        self.components = {}
        self.component_pairs = {}

        for id, component in self.threatmodel.component_library.components.items():
            self.components[id] = component.name.split(":")[-1] # Dirty hack
            for path in component.paths:
                if not path:
                    continue

                i = 0
                while i < len(path)-1:
                    source_component = path[i]
                    source_component_id = self.to_id(source_component)
                    destination_component = path[i+1]
                    destination_component_id = self.to_id(destination_component)

                    if not source_component_id in self.components:
                        self.components[source_component_id] = source_component
                    if not destination_component_id in self.components:
                        self.components[destination_component] = destination_component

                    if source_component_id not in self.component_pairs:
                        self.component_pairs[source_component_id] = {}
                    if destination_component_id not in self.component_pairs[source_component_id]:
                        self.component_pairs[source_component_id][destination_component_id] = 1
                    else:
                        self.component_pairs[source_component_id][destination_component_id] +=1
                    i += 1
            
                last_component = path[-1]
                last_component_id = self.to_id(last_component)

                if last_component_id not in self.components:
                    self.components[last_component_id] = last_component
                
                if last_component_id not in self.component_pairs:
                    self.component_pairs[last_component_id] = {}
                if id not in self.component_pairs[last_component_id]:
                    self.component_pairs[last_component_id][id] = 1
                else:
                    self.component_pairs[last_component_id][id] += 1


class MarkdownTable():
    def __init__(self):
        self.headers = []
        self.rows = []

    def add_headers(self, headers):
        self.headers = headers

    def add_row(self, row):
        self.rows.append(row)

class Markdown():
    def __init__(self):
        self.data = ""

    def add_h1(self, text):
        self.data += "# {}\n\n".format(text)

    def add_paragraph(self, text):
        self.data += "{}\n\n".format(text)

    def add_image(self, alt, file, title):
        self.data += "![{}]({} \"{}\")\n\n".format(alt, file, title)

    def add_table(self, table):
        self.data += "| {} |\n".format(" | ".join(table.headers))
        self.data += "| {} |\n".format(" | ".join("---" for x in table.headers))
        for row in table.rows:
            self.data += "| {} |\n".format(" | ".join(row))
        self.data += "\n\n"

    def code(self, text):
        return "`{}`".format(text)

    def code_block(self, text):
        return "```\n{}\n```".format(text)


class Graph():
    def __init__(self, title):
        self.dot = Digraph(comment=title, # TODO: Unhardcode
            engine='dot',
            format='png',
            graph_attr={
                'rankdir': 'LR'
            },
            node_attr={
                'fontcolor': '#2f3640',
                'shape': 'rect',
                'fontsize': '10',
                'fontname': 'Helvetica'
            },
            edge_attr={
                'fontcolor': '#2f3640',
                'fontsize': '10',
                'fontname': 'Helvetica'
            }
        )

class MarkdownReporter(Reporter):
        
    def generate(self):
        self.parse_component_paths()

        self.graph = Graph(self.project.name)
        self.report = Markdown()
        self.report.add_h1(self.project.name) # TODO: Unhardcode
        self.report.add_paragraph(self.project.description)

        self.report.add_h1("Diagram")
        self.report.add_image("Diagram", "ThreatModel.png", "Threat Model Diagram")

        # Tests
        tests_by_component_control = {}
        for test in self.threatmodel.tests:
            component_id = test.component
            control_id = test.control

            if component_id not in tests_by_component_control:
                tests_by_component_control[component_id] = {}
            if control_id not in tests_by_component_control[component_id]:
                tests_by_component_control[component_id][control_id] = test

        # Components
        for component_id, component in self.components.items():
            self.graph.dot.node(component_id, component, color='#3498db')

        for comp_a_id in self.component_pairs:
            for comp_b_id in self.component_pairs[comp_a_id]:
                self.graph.dot.edge(comp_a_id, comp_b_id, color='#a3cdf7')

        # Threats
        for threat_id, threat in self.threatmodel.threat_library.threats.items():
            self.graph.dot.node(threat_id, "Threat\n{}\n\n{}".format(threat_id, threat.name), color='#c0392b')

        # Controls
        for control_id, control in self.threatmodel.control_library.controls.items():
            self.graph.dot.node(control_id, "Control\n{}\n\n{}".format(control_id, control.name), color='#27ae60')

        self.report.add_h1("Threats")

        table = MarkdownTable()
        table.add_headers(["Type", "Component", "Threat", "Description", "Test", "Test File", "File", "Line", "Source"])

        # Exposes
        for exposure in self.threatmodel.exposures:
            exposure_id = self.random_id()
            threat_id = exposure.threat
            threat = self.threatmodel.threat_library.threats[threat_id]
            component_id = exposure.component
            component = self.threatmodel.component_library.components[component_id]

            self.graph.dot.node(exposure_id, "Exposure\n\n{}".format(exposure.details), color='#c0392b')
            #dot.edge(threat["name"], component["name"], color='#c0392b', concentrate='true')
            #dot.edge(exposure["exposure"], threat["name"], color='#c0392b', concentrate='true')
            self.graph.dot.edge(exposure_id, component_id, color='#c0392b', concentrate='true')
            self.graph.dot.edge(threat_id, exposure_id, color='#c0392b', concentrate='true')

            table.add_row([
                "Exposure",
                component.name,
                threat.name,
                exposure.details,
                "",
                "",
                exposure.source.filename,
                str(exposure.source.line),
                self.report.code(exposure.source.code)
            ])

        # Acceptances
        for acceptance in self.threatmodel.acceptances:
            acceptance_id = self.random_id()
            threat_id = acceptance.threat
            threat = self.threatmodel.threat_library.threats[threat_id]
            component_id = acceptance.component
            component = self.threatmodel.component_library.components[component_id]

            self.graph.dot.node(acceptance_id, "Accepts\n\n{}".format(acceptance.details), color='#c0392b')
            self.graph.dot.edge(threat_id, component_id, color='#c0392b', concentrate='true')
            self.graph.dot.edge(acceptance_id, threat_id, color='#c0392b', concentrate='true')

            table.add_row([
                "Acceptance",
                component.name,
                threat.name,
                acceptance.details,
                "",
                "",
                acceptance.source.filename,
                str(acceptance.source.line),
                self.report.code(acceptance.source.code)
            ])

        # Transfers
        for transfer in self.threatmodel.transfers:
            transfer_id = self.random_id()
            threat_id = transfer.threat
            threat = self.threatmodel.threat_library.threats[threat_id]
            source_id = transfer.source_component
            source = self.threatmodel.component_library.components[source_id]
            dest_id = transfer.destination_component
            dest = self.threatmodel.component_library.components[dest_id]

            self.graph.dot.node(transfer_id, "Transfer\n\n{}".format(transfer.details), color='#8e44ad')

            #dot.edge(source["name"], threat["name"], color='#f39c12', concentrate='true')
            #dot.edge(threat["name"], dest["name"], color='#e74c3c', concentrate='true')
            #dot.edge(transfer["transfer"], threat["name"], color='#8e44ad', concentrate='true')
            
            self.graph.dot.edge(source_id, transfer_id, color='#f39c12', concentrate='true')
            self.graph.dot.edge(transfer_id, dest_id, color='#e74c3c', concentrate='true')
            self.graph.dot.edge(threat_id, transfer_id, color='#8e44ad', concentrate='true')

            table.add_row(
                ["Transfer",
                "{} (from {})".format(dest.name, source.name),
                threat.name,
                transfer.details,
                "",
                "",
                transfer.source.filename,
                str(transfer.source.line),
                self.report.code(transfer.source.code)
            ])

        # Mitigations
        for mitigation in self.threatmodel.mitigations:
            component_id = mitigation.component
            component = self.threatmodel.component_library.components[component_id]
            threat_id = mitigation.threat
            threat = self.threatmodel.threat_library.threats[threat_id]
            control_id = mitigation.control
            control = self.threatmodel.control_library.controls[control_id]

            #dot.edge(threat["name"], component["name"], color='#f39c12', concentrate='true')
            #dot.edge(control["name"], threat["name"], color='#27ae60', concentrate='true')
            
            self.graph.dot.edge(control_id, component_id, color='#27ae60', concentrate='true')
            self.graph.dot.edge(threat_id, control_id, color='#f39c12', concentrate='true')

            if component_id in tests_by_component_control and control_id in tests_by_component_control[component_id]:
                test = tests_by_component_control[component_id][control_id]
                test_field = self.report.code(test.source.code)
                test_line = "{}:{}".format(test.source.filename, str(test.source.line))
            else:
                test_field = "None"
                test_line = ""

            table.add_row([
                "Mitigation",
                component.name,
                threat.name,
                control.name,
                test_field,
                test_line,
                mitigation.source.filename,
                str(mitigation.source.line),
                self.report.code(mitigation.source.code)
            ])

        # Create the threats table
        self.report.add_table(table)

        # Connects
        self.report.add_h1("Connections")
        table = MarkdownTable()
        table.add_headers(["Source", "Destination", "Description", "File", "Line", "Source"])

        for connection in self.threatmodel.connections:
            source_id = connection.source_component
            source = self.threatmodel.component_library.components[source_id]
            dest_id = connection.destination_component
            dest = self.threatmodel.component_library.components[dest_id]

            self.graph.dot.edge(source_id, dest_id, label=connection.details, concentrate='true')

            table.add_row([
                source.name,
                dest.name,
                connection.details,
                connection.source.filename,
                str(connection.source.line),
                self.report.code(connection.source.code)
            ])

        self.report.add_table(table)

        # Reviews
        self.report.add_h1("Reviews")
        table = MarkdownTable()
        table.add_headers(["Component", "Review", "File", "Line", "Source"])

        for review in self.threatmodel.reviews:
            review_id = self.random_id()
            component_id = review.component
            component = self.threatmodel.component_library.components[component_id]

            self.graph.dot.node(review_id, review.details)
            self.graph.dot.edge(review_id, component_id)

            table.add_row([
                component.name,
                review.details,
                review.source.filename,
                str(review.source.line),
                self.report.code(review.source.code)
            ])

        self.report.add_table(table)

        # Outputs
        self.graph.dot.render('ThreatModel.gv')
        data.write_file(self.report.data, "ThreatModel.md")



from setuptools import setup, find_packages

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()
    
setup(
    name='threatspec',
    description='threat modeling as code',
    long_description=long_description,
    long_description_content_type='text/markdown',
    version='0.2.5',
    license='MIT',
    author='Fraser Scott',
    author_email='fraser.scott@gmail.com',
    url='https://github.com/threatspec/threatspec',
    download_url='https://github.com/threatspec/threatspec/archive/v0.2.2-alpha.tar.gz',
    keywords=['threat modeling', 'cyber security', 'appsec'],
    packages=find_packages(),
    include_package_data=True,
    python_requires='>=3.5',
    install_requires=[
        'Click',
        'graphviz',
        'pyyaml',
        'jsonschema'
    ],
    setup_requires = [
        "pytest-runner"
    ],
    tests_require = [
        'pytest'
    ],
    entry_points='''
        [console_scripts]
        threatspec=threatspec.cli:cli
    ''',
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'Topic :: Security',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7'
    ]
)



import logging
logger = logging.getLogger(__name__)

import yaml, re
from threatspec.threatmodel import ThreatModel

class Parser():
    def __init__(self, threatmodel):

        self.threatmodel = threatmodel

        self.action_table = {}
        self.action_table["mitigates"] = self.threatmodel.add_mitigation
        self.action_table["accepts"] = self.threatmodel.add_acceptance
        self.action_table["transfers"] = self.threatmodel.add_transfer
        self.action_table["exposes"] = self.threatmodel.add_exposure
        self.action_table["connects"] = self.threatmodel.add_connection
        self.action_table["review"] = self.threatmodel.add_review
        self.action_table["tests"] = self.threatmodel.add_test

        self.comment_chars = ["//", "#"]

        self.patterns = {}
        self.patterns["mitigates"] = r'@mitigates (?P<component>.*?) against (?P<threat>.*?) with (?P<control>.*)'
        self.patterns["accepts"] = r'@accepts (?P<threat>.*?) to (?P<component>.*?) with (?P<details>.*)'
        self.patterns["transfers"] = r'@transfers (?P<threat>.*?) from (?P<source_component>.*?) to (?P<destination_component>.*?) with (?P<details>.*)'
        self.patterns["exposes"] = r'@exposes (?P<component>.*?) to (?P<threat>.*?) with (?P<details>.*)'
        self.patterns["connects"] = r'@connects (?P<source_component>.*?) (?P<direction>with|to) (?P<destination_component>.*?) with (?P<details>.*)'
        self.patterns["review"] = r'@review (?P<component>.*?) (?P<details>.*)'
        self.patterns["tests"] = r'@tests (?P<control>.*?) for (?P<component>.*)'


    def parse_annotation(self, annotation):
        for action in self.patterns.keys():
            if annotation.startswith("@"+action):
                data = {"action": action}
                pattern = self.patterns[action]
                m = re.match(pattern, annotation, re.M | re.I)
                if m:
                    data.update(m.groupdict())
                    return data
                else:
                    raise RuntimeError("Unable to match annotation '{}' with pattern '{}'".format(annotation, pattern))

    def run_action(self, data, source):
        action = data.pop("action")
        self.action_table[action](**data, source=source)

    def is_threatspec_line(self, line):
        for key in self.patterns.keys():
            if "@{}".format(key) in line:
                return True
        return False

class SourceFileParser(Parser):
    def parse_file(self, filename):
        try:
            with open(filename) as fh:
                lines = fh.readlines()
        except UnicodeDecodeError:
            return
        logger.debug("Parsing file {}".format(filename))
        current_line_index = 0

        while current_line_index < len(lines):
            current_line = lines[current_line_index].strip()
            if self.is_threatspec_line(current_line):
                next_line_index = current_line_index + 1
                while next_line_index < len(lines):
                    next_line = lines[next_line_index].strip()
                    if not self.is_threatspec_line(next_line):
                        logger.debug("Parsing line {}".format(current_line))
                        (data, source) = self.parse_line(current_line, next_line, filename,  current_line_index+1)
                        if data:
                            self.run_action(data, source)
                        break
                    next_line_index += 1
            current_line_index += 1

    def parse_comment_line(self, line):
        annotation = ""
        code = ""
        for commment_char in self.comment_chars: # TODO: Support files without comments
            if commment_char in line:
                parts = line.split(commment_char)
                annotation = parts[-1].strip()
                code = "".join(parts[0:-1]).strip()
                break
        return (annotation, code)

    def parse_line(self, line, next_line, filename, line_no):
        (annotation, code) = self.parse_comment_line(line)
        if annotation == "":
            return (None, None)
        if code == "":
            code = next_line

        data = self.parse_annotation(annotation)
        if not data:
            return (None, None)
        source = {
            "annotation": annotation,
            "code": code,
            "filename": filename,
            "line": line_no
        }
        return (data, source)

"""
class YamlFileParser(Parser):
    def parse_data(self, data, parent, filename):
        if isinstance(data, dict):
            for k, v in data.items():
                if k == "x-threatspec":
                    for action in self.patterns.keys():
                        if v.startswith("@"+action):
                            self.parse(action, v, str(parent), filename, 0)
                else:
                    self.parse_data(v, data, filename)
        elif isinstance(data, list):
            for v in data:
                self.parse_data(v, data, filename)

    def parse_file(self, filename):
        with open(filename) as fh:
            file_data = yaml.load(fh, Loader=yaml.SafeLoader)
            self.parse_data(file_data, {}, filename)
"""


from typing import List, Dict, Any
from threatspec import data

class Project():
    def __init__(self, name: str = "", description: str = ""):
        self.name = name
        self.description = description

class Path:
    def __init__(self, obj):
        self.path = ""
        self.ignore = ""

        if isinstance(obj, str):
            self.path = obj
        elif isinstance(obj, dict):
            if "path" not in obj:
                raise ValueError("path key missing from path")
            self.path = obj["path"]
            if "ignore" in obj:
                if isinstance(obj["ignore"], str):
                    self.ignore = [obj["ignore"]]
                elif isinstance(obj["ignore"], list):
                    self.ignore = obj["ignore"]
                else:
                    raise TypeError("ignore must be a string or list")

class Config:
    def __init__(self, project: Project = Project(), paths: List[Path] = []):
        self.project = project
        self.paths = paths

    def load(self, data):
        self.project = Project(data["project"]["name"], data["project"]["description"])
        for path in data["paths"]:
            self.paths.append(Path(path))